import datetime
import json
import typing as t
import uuid
from dataclasses import dataclass, field

import requests

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseDestinationConnector,
    BaseIngestDoc,
    WriteConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import flatten_dict

BASE_URL = "https://api.vectara.io/v1"


@dataclass
class VectaraAccessConfig(AccessConfig):
    oauth_client_id: str = enhanced_field(sensitive=True)
    oauth_secret: str = enhanced_field(sensitive=True)


@dataclass
class SimpleVectaraConfig(BaseConnectorConfig):
    access_config: VectaraAccessConfig
    customer_id: str
    corpus_name: t.Optional[str] = None
    corpus_id: t.Optional[str] = None
    token_url: str = "https://vectara-prod-{}.auth.us-west-2.amazoncognito.com/oauth2/token"


@dataclass
class VectaraDestinationConnector(BaseDestinationConnector):
    write_config: WriteConfig
    connector_config: SimpleVectaraConfig
    _jwt_token: t.Optional[str] = field(init=False, default=None)
    _jwt_token_expires_ts: t.Optional[float] = field(init=False, default=None)

    @property
    def jwt_token(self):
        if (
            not self._jwt_token
            or self._jwt_token_expires_ts - datetime.datetime.now().timestamp() <= 60
        ):
            self._jwt_token = self._get_jwt_token()
        return self._jwt_token

    @DestinationConnectionError.wrap
    def vectara(self):
        """
        Check the connection for Vectara and validate corpus exists.
        - If more than one corpus with the same name exists - then return a message
        - If exactly one corpus exists with this name - use it.
        - If does not exist - create it.
        """
        try:
            # Get token if not already set
            self.jwt_token

            list_corpora_response = self._request(
                endpoint="list-corpora",
                data={"numResults": 1, "filter": self.connector_config.corpus_name},
            )

            possible_corpora_ids_names_map = {
                corpus.get("id"): corpus.get("name")
                for corpus in list_corpora_response.get("corpus")
                if corpus.get("name") == self.connector_config.corpus_name
            }

            if len(possible_corpora_ids_names_map) > 1:
                return f"Multiple Corpora exist with name {self.connector_config.corpus_name}"
            if len(possible_corpora_ids_names_map) == 1:
                self.connector_config.corpus_id = list(possible_corpora_ids_names_map.keys())[0]
            else:
                data = {
                    "corpus": {
                        "name": self.connector_config.corpus_name,
                    }
                }
                create_corpus_response = self._request(endpoint="create-corpus", data=data)
                self.connector_config.corpus_id = create_corpus_response.get("corpusId")

        except Exception as e:
            logger.error(f"failed to create Vectara connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to create Vectara connection: {e}")

    def initialize(self):
        self.vectara()

    def _request(
        self,
        endpoint: str,
        http_method: str = "POST",
        params: t.Mapping[str, t.Any] = None,
        data: t.Mapping[str, t.Any] = None,
    ):
        url = f"{BASE_URL}/{endpoint}"

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {self.jwt_token}",
            "customer-id": self.connector_config.customer_id,
            "X-source": "unstructured",
        }

        response = requests.request(
            method=http_method, url=url, headers=headers, params=params, data=json.dumps(data)
        )
        response.raise_for_status()
        return response.json()

    # Get Oauth2 JWT token
    def _get_jwt_token(self):
        """Connect to the server and get a JWT token."""
        token_endpoint = self.connector_config.token_url.format(self.connector_config.customer_id)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {
            "grant_type": "client_credentials",
            "client_id": self.connector_config.access_config.oauth_client_id,
            "client_secret": self.connector_config.access_config.oauth_secret,
        }

        response = requests.request(method="POST", url=token_endpoint, headers=headers, data=data)
        response.raise_for_status()
        response_json = response.json()

        request_time = datetime.datetime.now().timestamp()
        self._jwt_token_expires_ts = request_time + response_json.get("expires_in")

        return response_json.get("access_token")

    @DestinationConnectionError.wrap
    def check_connection(self):
        try:
            self.vectara()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def _delete_doc(self, doc_id: str) -> None:
        """
        Delete a document from the Vectara corpus.

        Args:
            url (str): URL of the page to delete.
            doc_id (str): ID of the document to delete.
        """
        body = {
            "customer_id": self.connector_config.customer_id,
            "corpus_id": self.connector_config.corpus_id,
            "document_id": doc_id,
        }
        self._request(endpoint="delete-doc", data=body)

    def _index_document(self, document: t.Dict[str, t.Any]) -> None:
        """
        Index a document (by uploading it to the Vectara corpus) from the document dictionary
        """
        body = {
            "customer_id": self.connector_config.customer_id,
            "corpus_id": self.connector_config.corpus_id,
            "document": document,
        }

        try:
            result = self._request(endpoint="index", data=body, http_method="POST")
        except Exception as e:
            logger.info(f"Exception {e} while indexing document {document['documentId']}")
            return

        if (
            "status" in result
            and result["status"]
            and (
                "ALREADY_EXISTS" in result["status"]["code"]
                or (
                    "CONFLICT" in result["status"]["code"]
                    and "Indexing doesn't support updating documents"
                    in result["status"]["statusDetail"]
                )
            )
        ):
            logger.info(f"Document {document['documentId']} already exists, re-indexing")
            self._delete_doc(document["documentId"])
            result = self._request(endpoint="index", data=body, http_method="POST")
            return

        if "status" in result and result["status"] and "OK" in result["status"]["code"]:
            logger.info(f"Indexing document {document['documentId']} succeeded")
        else:
            logger.info(f"Indexing document {document['documentId']} failed, response = {result}")

    def write_dict(self, *args, docs_list: t.List[t.Dict[str, t.Any]], **kwargs) -> None:
        logger.info(f"Inserting / updating {len(docs_list)} documents to Vectara ")
        for vdoc in docs_list:
            self._index_document(vdoc)

    def write(self, docs: t.List[BaseIngestDoc]) -> None:
        docs_list: t.Dict[t.Dict[str, t.Any]] = []

        def get_metadata(element) -> t.Dict[str, t.Any]:
            """
            Select which meta-data fields to include and optionaly map them to a new new.
            remove the "metadata-" prefix from the keys
            """
            metadata_map = {
                "page_number": "page_number",
                "data_source-url": "url",
                "filename": "filename",
                "filetype": "filetype",
                "last_modified": "last_modified",
            }
            md = flatten_dict(element, separator="-", flatten_lists=True)
            md = {k.replace("metadata-", ""): v for k, v in md.items()}
            md = {metadata_map[k]: v for k, v in md.items() if k in metadata_map}
            return md

        for doc in docs:
            local_path = doc._output_filename
            with open(local_path) as json_file:
                dict_content = json.load(json_file)
                vdoc = {
                    "documentId": str(uuid.uuid4()),
                    "title": dict_content[0].get("metadata", {}).get("data_source", {}).get("url"),
                    "section": [
                        {
                            "text": element.pop("text", None),
                            "metadataJson": json.dumps(get_metadata(element)),
                        }
                        for element in dict_content
                    ],
                }
                logger.info(
                    f"Extending {len(vdoc)} json elements from content in {local_path}",
                )
                docs_list.append(vdoc)
        self.write_dict(docs_list=docs_list)



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.pinecone import PineconeWriteConfig, SimplePineconeConfig


@dataclass
class PineconeWriter(Writer):
    write_config: "PineconeWriteConfig"
    connector_config: "SimplePineconeConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.pinecone import (
            PineconeDestinationConnector,
        )

        return PineconeDestinationConnector



from dataclasses import dataclass

import click
from unstructured.chunking import CHUNK_MAX_CHARS_DEFAULT, CHUNK_MULTI_PAGE_DEFAULT

from unstructured_ingest.v2.cli.interfaces import CliConfig


@dataclass
class ChunkerCliConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--chunking-strategy"],
                type=str,
                default=None,
                help="The rule-set to use to form chunks. Omit to disable chunking.",
            ),
            click.Option(
                ["--chunk-combine-text-under-n-chars"],
                type=int,
                help=(
                    "Combine consecutive chunks when the first does not exceed this length and"
                    " the second will fit without exceeding the hard-maximum length. Only"
                    " operative for 'by_title' chunking-strategy."
                ),
            ),
            click.Option(
                ["--chunk-include-orig-elements/--chunk-no-include-orig-elements"],
                is_flag=True,
                default=True,
                help=(
                    "When chunking, add the original elements consolidated to form each chunk to"
                    " `.metadata.orig_elements` on that chunk."
                ),
            ),
            click.Option(
                ["--chunk-max-characters"],
                type=int,
                default=CHUNK_MAX_CHARS_DEFAULT,
                show_default=True,
                help=(
                    "Hard maximum chunk length. No chunk will exceed this length. An oversized"
                    " element will be divided by text-splitting to fit this window."
                ),
            ),
            click.Option(
                ["--chunk-multipage-sections/--chunk-no-multipage-sections"],
                is_flag=True,
                default=CHUNK_MULTI_PAGE_DEFAULT,
                help=(
                    "Ignore page boundaries when chunking such that elements from two different"
                    " pages can appear in the same chunk. Only operative for 'by_title'"
                    " chunking-strategy."
                ),
            ),
            click.Option(
                ["--chunk-new-after-n-chars"],
                type=int,
                help=(
                    "Soft-maximum chunk length. Another element will not be added to a chunk of"
                    " this length even when it would fit without exceeding the hard-maximum"
                    " length."
                ),
            ),
            click.Option(
                ["--chunk-overlap"],
                type=int,
                default=0,
                show_default=True,
                help=(
                    "Prefix chunk text with last overlap=N characters of prior chunk. Only"
                    " applies to oversized chunks divided by text-splitting. To apply overlap to"
                    " non-oversized chunks use the --overlap-all option."
                ),
            ),
            click.Option(
                ["--chunk-overlap-all"],
                is_flag=True,
                default=False,
                help=(
                    "Apply overlap to chunks formed from whole elements as well as those formed"
                    " by text-splitting oversized elements. Overlap length is take from --overlap"
                    " option value."
                ),
            ),
        ]
        return options



import typing as t
from dataclasses import dataclass, field
from pathlib import Path

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.string_and_date_utils import ensure_isoformat_datetime

if t.TYPE_CHECKING:
    from office365.graph_client import GraphClient
    from office365.onedrive.driveitems.driveItem import DriveItem
MAX_MB_SIZE = 512_000_000


@dataclass
class OneDriveAccessConfig(AccessConfig):
    client_credential: str = enhanced_field(repr=False, sensitive=True, overload_name="client_cred")


@dataclass
class SimpleOneDriveConfig(BaseConnectorConfig):
    access_config: OneDriveAccessConfig
    client_id: str
    user_pname: str
    tenant: str = field(repr=False)
    authority_url: t.Optional[str] = field(repr=False, default="https://login.microsoftonline.com")
    path: t.Optional[str] = field(default="")
    recursive: bool = False

    def __post_init__(self):
        if not (self.client_id and self.access_config.client_credential and self.user_pname):
            raise ValueError(
                "Please provide all the following mandatory values:"
                "\n-ms-client_id\n-ms-client_cred\n-ms-user-pname",
            )
        self.token_factory = self._acquire_token

    @SourceConnectionError.wrap
    @requires_dependencies(["msal"])
    def _acquire_token(self):
        from msal import ConfidentialClientApplication

        try:
            app = ConfidentialClientApplication(
                authority=f"{self.authority_url}/{self.tenant}",
                client_id=self.client_id,
                client_credential=self.access_config.client_credential,
            )
            token = app.acquire_token_for_client(scopes=["https://graph.microsoft.com/.default"])
        except ValueError as exc:
            logger.error("Couldn't set up credentials for OneDrive")
            raise exc
        return token


@dataclass
class OneDriveIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    connector_config: SimpleOneDriveConfig
    file_name: str
    file_path: str
    registry_name: str = "onedrive"

    def __post_init__(self):
        self.ext = Path(self.file_name).suffix
        if not self.ext:
            raise ValueError("Unsupported file without extension.")

        self.server_relative_path = self.file_path + "/" + self.file_name
        self._set_download_paths()

    def _set_download_paths(self) -> None:
        """Parses the folder structure from the source and creates the download and output paths"""
        download_path = Path(f"{self.read_config.download_dir}")
        output_path = Path(f"{self.processor_config.output_dir}")

        if parent_path := self.file_path:
            download_path = (
                download_path if parent_path == "" else (download_path / parent_path).resolve()
            )
            output_path = (
                output_path if parent_path == "" else (output_path / parent_path).resolve()
            )

        self.download_dir = download_path
        self.download_filepath = (download_path / self.file_name).resolve()
        output_filename = output_filename = self.file_name + ".json"
        self.output_dir = output_path
        self.output_filepath = (output_path / output_filename).resolve()

    @property
    def filename(self):
        return Path(self.download_filepath).resolve()

    @property
    def _output_filename(self):
        return Path(self.output_filepath).resolve()

    @property
    def record_locator(self) -> t.Optional[t.Dict[str, t.Any]]:
        return {
            "user_pname": self.connector_config.user_pname,
            "server_relative_path": self.server_relative_path,
        }

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(["office365"], extras="onedrive")
    def _fetch_file(self):
        from office365.graph_client import GraphClient

        client = GraphClient(self.connector_config.token_factory)
        root = client.users[self.connector_config.user_pname].drive.get().execute_query().root
        file = root.get_by_path(self.server_relative_path).get().execute_query()
        return file

    def update_source_metadata(self, **kwargs):
        file = kwargs.get("file", self._fetch_file())
        if file is None:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return

        version = None
        if (n_versions := len(file.versions)) > 0:
            version = file.versions[n_versions - 1].properties.get("id", None)

        self.source_metadata = SourceMetadata(
            date_created=ensure_isoformat_datetime(timestamp=file.created_datetime),
            date_modified=ensure_isoformat_datetime(timestamp=file.last_modified_datetime),
            version=version,
            source_url=file.parent_reference.path + "/" + self.file_name,
            exists=True,
        )

    @SourceConnectionError.wrap
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        file = self._fetch_file()
        self.update_source_metadata(file=file)
        if file is None:
            raise ValueError(
                f"Failed to retrieve file {self.file_path}/{self.file_name}",
            )

        fsize = file.get_property("size", 0)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        if not self.download_dir.is_dir():
            logger.debug(f"Creating directory: {self.download_dir}")
            self.download_dir.mkdir(parents=True, exist_ok=True)

        if fsize > MAX_MB_SIZE:
            logger.info(f"Downloading file with size: {fsize} bytes in chunks")
            with self.filename.open(mode="wb") as f:
                file.download_session(f, chunk_size=1024 * 1024 * 100).execute_query()
        else:
            with self.filename.open(mode="wb") as f:
                file.download(f).execute_query()
        logger.info(f"File downloaded: {self.filename}")
        return


@dataclass
class OneDriveSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    connector_config: SimpleOneDriveConfig
    _client: t.Optional["GraphClient"] = field(init=False, default=None)

    @property
    def client(self) -> "GraphClient":
        from office365.graph_client import GraphClient

        if self._client is None:
            self._client = GraphClient(self.connector_config.token_factory)
        return self._client

    @requires_dependencies(["office365"], extras="onedrive")
    def initialize(self):
        _ = self.client

    @requires_dependencies(["office365"], extras="onedrive")
    def check_connection(self):
        try:
            token_resp: dict = self.connector_config.token_factory()
            if error := token_resp.get("error"):
                raise SourceConnectionError(
                    "{} ({})".format(error, token_resp.get("error_description"))
                )
            _ = self.client
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def _list_objects(self, folder, recursive) -> t.List["DriveItem"]:
        drive_items = folder.children.get().execute_query()
        files = [d for d in drive_items if d.is_file]
        if not recursive:
            return files
        folders = [d for d in drive_items if d.is_folder]
        for f in folders:
            files += self._list_objects(f, recursive)
        return files

    def _gen_ingest_doc(self, file: "DriveItem") -> OneDriveIngestDoc:
        file_path = file.parent_reference.path.split(":")[-1]
        file_path = file_path[1:] if file_path[0] == "/" else file_path
        return OneDriveIngestDoc(
            connector_config=self.connector_config,
            processor_config=self.processor_config,
            read_config=self.read_config,
            file_name=file.name,
            file_path=file_path,
        )

    def get_ingest_docs(self):
        root = self.client.users[self.connector_config.user_pname].drive.get().execute_query().root
        if fpath := self.connector_config.path:
            root = root.get_by_path(fpath).get().execute_query()
            if root is None or not root.is_folder:
                raise ValueError(f"Unable to find directory, given: {fpath}")
        files = self._list_objects(root, self.connector_config.recursive)
        return [self._gen_ingest_doc(f) for f in files]
